#ifndef _DMA_LST_H
#define _DMA_LST_H

#include <stdio.h>
//#include <spu_mfcio.h>

//#define DEBUG  // DMA related info 
//#define DEBUG2 //other debug info  

#ifdef DEBUG
  #define MYOUT_DMA(format,args...) fprintf(stderr, format, ## args)
#else
  #define MYOUT_DMA(format,args...) 
#endif

#ifdef DEBUG2
  #define MYOUT(format,args...) fprintf(stderr, format, ## args)
#else
  #define MYOUT(format,args...) 
#endif

/* UP round number to multiple of 16 */ 
#define RND16U(num) ((((unsigned int)(num))+0xf)&~0xf)
/* remains of UP round 16 */
#define RND16U_REM(num) (0xf-(((unsigned int)(num))&0xf))
/* DOWN round number to multiple of 16 */ 
#define RND16D(num) (((unsigned int)(num))&~0xf)
/* remains of DOWN round 16 */ 
#define RND16D_REM(num) (((unsigned int)(num))&0xf)

/* DMA_LIST_INIT */
#define DL_INIT(lst_num,p)  \
  int _dl_ind##p = 0; \
  dma_lst_elem_t* p = x264_malloc(lst_num*sizeof(dma_lst_elem_t))

/* DMA_LST_EXIT */
#define DL_EXIT(p) x264_free(p)

/* DMA_LST_CLEAR */
#define DL_CLR(p) _dl_ind##p = 0

/* DMA_LST_POP */
#define DL_POP(p) _dl_ind##p--
/* DMA_LST_ELEM */
#define DL_ELE(member,p) DL_ELEP(member,sizeof((member)[0]),p)

#define DL_ELEP(member,sz,p) { \
  MYOUT_DMA("%s[%d]: DMA_LIST_ELE[%d]: "#member" = %x, "#sz" = %d\n", \
        __func__, __LINE__, _dl_ind##p, (unsigned int)member, (sz)); \
  p[_dl_ind##p].ea_low = (unsigned int)(member); \
  p[_dl_ind##p].size.all = (unsigned int)(sz); \
  _dl_ind##p++; \
}
/* DMA_LST_GET */
#define DL_GET(dst,tagid,p) { \
  MYOUT_DMA("%s[%d]: DMA_LIST_GET: "#dst" = %x\n", __func__, __LINE__, (unsigned int)dst); \
  spu_mfcdma32(dst, (unsigned int) &p[0], \
              _dl_ind##p * sizeof(dma_lst_elem_t), tagid, MFC_GETL_CMD); }
#define DL_PUT(src,tagid,p) { \
  MYOUT_DMA("%s[%d]: DMA_LIST_PUT: "#src" = %x\n", __func__, __LINE__, (unsigned int)src); \
  spu_mfcdma32(src, (unsigned int) &p[0], \
              _dl_ind##p * sizeof(dma_lst_elem_t), tagid, MFC_PUTL_CMD); }
/* DMA WAIT on TAG */
#define DL_WAIT(tagid) \
  mfc_write_tag_mask(1<<(tagid)); \
  mfc_read_tag_status_all();
#define DL_WAIT2(tagid1, tagid2) \
  mfc_write_tag_mask((1<<(tagid1))|(1<<(tagid2))); \
  mfc_read_tag_status_all();
#define DL_WAIT3(tagid1, tagid2, tagid3) \
  mfc_write_tag_mask((1<<(tagid1))|(1<<(tagid2))|(1<<(tagid3))); \
  mfc_read_tag_status_all();

/* DMA SINGLE GET */
#define DS_GET(ls,ea,size,tag) \
  MYOUT_DMA("%s[%d]: MFC_GET: "#ls" = %x, "#ea" = %x, "#size" = %d\n", \
        __func__, __LINE__, (unsigned int)ls, (unsigned int)ea, (size)); \
  mfc_get(ls,(unsigned int)ea,(size),tag,0,0)
/* DMA SINGLE PUT */
#define DS_PUT(ls,ea,size,tag) \
  MYOUT_DMA("%s[%d]: MFC_PUT: "#ls" = %x, "#ea" = %x, "#size" = %d\n", \
        __func__, __LINE__, (unsigned int)ls, (unsigned int)ea, (size)); \
  mfc_put(ls,(unsigned int)ea,(size),tag,0,0)

#define DS_WAIT(t) DL_WAIT(t)
#define DS_WAIT2(t1,t2) DL_WAIT2(t1,t2)
#define DS_WAIT3(t1,t2,t3) DL_WAIT3(t1,t2,t3)

/* DMA LIST for LARGE buffer */
#define DL_GET_LARGE(ls,ea,size,tag) { \
  int _sz = (size); \
  DL_INIT(_sz/(16*1024)+1,_LARGE_); \
  while (_sz > 16*1024 ){ \
    DL_ELEP(ea,16*1024,_LARGE_);\
    _sz -= 16*1024; \
  } \
  if (_sz > 0){ \
    DL_ELEP(ea,_sz,_LARGE_); \
  } \
  DL_GET(ls,tag,_LARGE_); \
  DL_WAIT(tag); \
  DL_EXIT(_LARGE_); \
}

#define DL_PUT_LARGE(ls,ea,size,tag) { \
  int _sz = (size); \
  DL_INIT(_sz/(16*1024)+1,_LARGE_); \
  while (_sz > 16*1024 ){ \
    DL_ELEP(ea,16*1024,_LARGE_);\
    _sz -= 16*1024; \
  } \
  if (_sz > 0){ \
    DL_ELEP(ea,_sz,_LARGE_); \
  } \
  DL_PUT(ls,tag,_LARGE_); \
  DL_WAIT(tag); \
  DL_EXIT(_LARGE_); \
}

/* compiler directives */
#define UNLIKE(a) __builtin_expect((a),0)
#define LIKE(a) __builtin_expect((a),1)
    
typedef struct dma_lst_elem {
  union {
    unsigned int all;
    struct {
      unsigned nbytes: 31;
      unsigned stall: 1;
    }bits;
  }size;
  unsigned int ea_low;
}dma_lst_elem_t;

#endif

