#include <me_spu.h>
#include "dma_list.h"

static const int subpel_iterations[][4] =
   {{0,0,0,0},
    {1,1,0,0},
    {0,1,1,0},
    {0,2,1,0},
    {0,2,1,1},
    {0,2,1,2},
    {0,0,2,2},
    {0,0,2,2},
    {0,0,4,10},
    {0,0,4,10}};

static const int mod6m1[8] = {5,0,1,2,3,4,5,0};
static const int hex2[8][2] = {{-1,-2}, {-2,0}, {-1,2}, {1,2}, {2,0}, {1,-2}, {-1,-2}, {-2,0}};

void *x264_malloc( int i_size )
{
#ifdef SYS_MACOSX
    /* Mac OS X always returns 16 bytes aligned memory */
    return malloc( i_size );
//#elif defined( HAVE_MALLOC_H )
    //return memalign( 16, i_size );
#else
    uint8_t * buf;
    uint8_t * align_buf;
    buf = (uint8_t *) malloc( i_size + 15 + sizeof( void ** ) +
              sizeof( int ) );
    align_buf = buf + 15 + sizeof( void ** ) + sizeof( int );
    align_buf -= (long) align_buf & 15;
    *( (void **) ( align_buf - sizeof( void ** ) ) ) = buf;
    *( (int *) ( align_buf - sizeof( void ** ) - sizeof( int ) ) ) = i_size;
    return align_buf;
#endif
}

void x264_free( void *p )
{
    if( p )
    {
//#if defined( HAVE_MALLOC_H ) || defined( SYS_MACOSX )
//        free( p );
//#else
        free( *( ( ( void **) p ) - 1 ) );
//#endif
    }
}

/*---DMA variables---*/
volatile x264_t h ALIGN16;
x264_me_t m ALIGN16;
int16_t mvc[9][2];
int i_mvc;
int *p_halfpel_thresh;
int i_halfpel_thresh;
/*-------------------*/

unsigned int sad_time = 0;


void dma_transfer(void *ls, unsigned int ea, int bytes, int dma_tag)
{
    unsigned int startaddress = ea - ea % 16;
    unsigned int transfersize = (ea + bytes - startaddress);
    void *buffer;
 
    if(transfersize > 16)
        transfersize = ((transfersize - 1) / 16 + 1) * 16;
    else
    {
        int i = 1;
        while(i < transfersize)
            i <<= 1;
        transfersize = i;
    }
    
    buffer = x264_malloc(transfersize);
    
    DMA_GET(buffer, startaddress, transfersize, dma_tag);
    DMA_WAIT(dma_tag);
    
    memcpy(ls, buffer + ea - startaddress, bytes);
    
    x264_free(buffer);    
}

int16_t get_int16_from_ppu(const int16_t *ppu_ptr)
{
    int16_t n;
    dma_transfer(&n, ppu_ptr, sizeof(int16_t), 0);
    return n;
}

void x264_me_search_ref( x264_t *ppu_h, x264_me_t *ppu_m, int16_t (*ppu_mvc)[2], int i_mvc, int *ppu_p_halfpel_thresh, int num )
{
    
    //printf("SPU %d starting me_search_ref()\n", num);
    //---- DMA transfers------------//
    DL_GET_LARGE(&h,(unsigned int)(ppu_h),sizeof(x264_t),0);
    dma_transfer(&m, (unsigned int)ppu_m, sizeof(x264_me_t), 0);
    dma_transfer(mvc, (unsigned int)&ppu_mvc[0][0], sizeof(int16_t)*9*2, 0);

    if(ppu_p_halfpel_thresh != NULL)
        dma_transfer(&i_halfpel_thresh, (unsigned int)ppu_p_halfpel_thresh, sizeof(int), 0);
    //------------------------------------------------------------------------//

    sad_time = 0;

    //printf("THREAD %3x  FRAME %3d                 SPU %d  x264_me_search_ref(): STARTING\n", (h.thread_handle>>16)&0x0FFF, h.i_frame, num);
    
    const int bw = x264_pixel_size[m.i_pixel].w;
    const int bh = x264_pixel_size[m.i_pixel].h;
    const int i_pixel = m.i_pixel;
    int i_me_range = h.param.analyse.i_me_range;
    int bmx, bmy, bcost;
    int bpred_mx = 0, bpred_my = 0, bpred_cost = COST_MAX;
    int omx, omy, pmx, pmy;

    uint8_t *p_fref = m.p_fref[0];
    DECLARE_ALIGNED_16( uint8_t pix[16*16] );

    int i, j;
    int dir;
    int costs[6];

    int mv_x_min = h.mb.mv_min_fpel[0];
    int mv_y_min = h.mb.mv_min_fpel[1];
    int mv_x_max = h.mb.mv_max_fpel[0];
    int mv_y_max = h.mb.mv_max_fpel[1];

#define CHECK_MVRANGE(mx,my) ( mx >= mv_x_min && mx <= mv_x_max && my >= mv_y_min && my <= mv_y_max )

    const int16_t *p_cost_mvx = m.p_cost_mv - m.mvp[0];
    const int16_t *p_cost_mvy = m.p_cost_mv - m.mvp[1];
    int16_t cost_mvx, cost_mvy;
 
    bmx = x264_clip3( m.mvp[0], mv_x_min*4, mv_x_max*4 );
    bmy = x264_clip3( m.mvp[1], mv_y_min*4, mv_y_max*4 );
    pmx = ( bmx + 2 ) >> 2;
    pmy = ( bmy + 2 ) >> 2;
    bcost = COST_MAX;
    /* try extra predictors if provided */
    //printf("IF %d\n", num);
    if( h.mb.i_subpel_refine >= 3 )
    {
        uint32_t bmv = pack16to32_mask(bmx,bmy);
        COST_MV_HPEL( bmx, bmy );
        for( i = 0; i < i_mvc; i++ )
        {
            if( *(uint32_t*)mvc[i] && (bmv - *(uint32_t*)mvc[i]) )
            {
                int mx = x264_clip3( mvc[i][0], mv_x_min*4, mv_x_max*4 );
                int my = x264_clip3( mvc[i][1], mv_y_min*4, mv_y_max*4 );
                COST_MV_HPEL( mx, my );
            }
        }
        bmx = ( bpred_mx + 2 ) >> 2;
        bmy = ( bpred_my + 2 ) >> 2;
        COST_MV( bmx, bmy );
        //printf("ENDIF %d\n", num);
    }
    else
    {
        /* check the MVP */
        COST_MV( pmx, pmy );
        /* Because we are rounding the predicted motion vector to fullpel, there will be
         * an extra MV cost in 15 out of 16 cases.  However, when the predicted MV is
         * chosen as the best predictor, it is often the case that the subpel search will
         * result in a vector at or next to the predicted motion vector.  Therefore, it is
         * sensible to remove the cost of the MV from the rounded MVP to avoid unfairly
         * biasing against use of the predicted motion vector. */
        bcost -= get_int16_from_ppu(&p_cost_mvx[(pmx)<<2]) + get_int16_from_ppu(&p_cost_mvy[(pmy)<<2]);
        for( i = 0; i < i_mvc; i++ )
        {
            int mx = (mvc[i][0] + 2) >> 2;
            int my = (mvc[i][1] + 2) >> 2;
            if( (mx | my) && ((mx-bmx) | (my-bmy)) )
            {
                mx = x264_clip3( mx, mv_x_min, mv_x_max );
                my = x264_clip3( my, mv_y_min, mv_y_max );
                COST_MV( mx, my );
            }
        }
    }
    COST_MV( 0, 0 );
    
    switch( h.mb.i_me_method )
    {
    case X264_ME_DIA:
        printf("Unsupported search mode X264_ME_DIA in me_search_ref()\n");
        break;
    case X264_ME_HEX:
        me_hex2:
        /* hexagon search, radius 2 */
        for( i = 0; i < i_me_range/2; i++ )
        {
            omx = bmx; omy = bmy;
            COST_MV( omx-2, omy   );
            COST_MV( omx-1, omy+2 );
            COST_MV( omx+1, omy+2 );
            COST_MV( omx+2, omy   );
            COST_MV( omx+1, omy-2 );
            COST_MV( omx-1, omy-2 );
            if( bmx == omx && bmy == omy )
                break;
            if( !CHECK_MVRANGE(bmx, bmy) )
                break;
        }

        /* square refine */
        omx = bmx; omy = bmy;
        COST_MV_X4(  0,-1,  0,1, -1,0, 1,0 );
        COST_MV_X4( -1,-1, -1,1, 1,-1, 1,1 );
        break;
    
    case X264_ME_UMH:
        printf("Unsupported search mode X264_ME_UMH in me_search_ref()\n");
        break;
    case X264_ME_ESA:
    case X264_ME_TESA:
        printf("Unsupported search mode X264_ME_ESA or X264_ME_TESA in me_search_ref()\n");
        break;
    default:
        printf("Unsupported search mode in me_search_ref()\n");
        break;
    }
    
    /* -> qpel mv */
    if( bpred_cost < bcost )
    {
        m.mv[0] = bpred_mx;
        m.mv[1] = bpred_my;
        m.cost = bpred_cost;
    }
    else
    {
        m.mv[0] = bmx << 2;
        m.mv[1] = bmy << 2;
        m.cost = bcost;
    }

    /* compute the real cost */
    m.cost_mv = get_int16_from_ppu(&p_cost_mvx[ m.mv[0] ]) + get_int16_from_ppu(&p_cost_mvy[ m.mv[1] ]);
    if( bmx == pmx && bmy == pmy && h.mb.i_subpel_refine < 3 )
        m.cost += m.cost_mv;
    DMA_PUT(&m, (unsigned int)ppu_m, sizeof(x264_me_t), 0);
    DMA_WAIT(0);
    //printf("SPU %d finishing me_search_ref()\n", num);
    //printf("THREAD %3x  FRAME %3d                 SPU %d  x264_me_search_ref(): FINISHED\n", (h.thread_handle>>16)&0x0FFF, h.i_frame, num); 
    //printf("The total SAD time takes: %u cycles\n", sad_time);
      
}

static void refine_subpel( x264_t *h, x264_me_t *m, int hpel_iters, int qpel_iters, int *p_halfpel_thresh, int b_refine_qpel )
{
    const int bw = x264_pixel_size[m->i_pixel].w;
    const int bh = x264_pixel_size[m->i_pixel].h;
    const int16_t *p_cost_mvx = m->p_cost_mv - m->mvp[0];
    const int16_t *p_cost_mvy = m->p_cost_mv - m->mvp[1];
    const int i_pixel = m->i_pixel;
    const int b_chroma_me = h->mb.b_chroma_me && i_pixel <= PIXEL_8x8;

    DECLARE_ALIGNED_16( uint8_t pix[2][32*18] ); // really 17x17, but round up for alignment
    int omx, omy;
    int i;

    int bmx = m->mv[0];
    int bmy = m->mv[1];
    int bcost = m->cost;
    int odir = -1, bdir;

    /* try the subpel component of the predicted mv */
    if( hpel_iters && h->mb.i_subpel_refine < 3 )
    {
        int mx = x264_clip3( m->mvp[0], h->mb.mv_min_spel[0], h->mb.mv_max_spel[0] );
        int my = x264_clip3( m->mvp[1], h->mb.mv_min_spel[1], h->mb.mv_max_spel[1] );
        if( (mx-bmx)|(my-bmy) )
            COST_MV_SAD( mx, my );
    }

    /* halfpel diamond search */
    for( i = hpel_iters; i > 0; i-- )
    {
        int omx = bmx, omy = bmy;
        int costs[4];
        int stride = 32; // candidates are either all hpel or all qpel, so one stride is enough
        uint8_t *src0, *src1, *src2, *src3;
        src0 = get_ref( pix[0], &stride, m->p_fref, m->i_stride[0], omx, omy-2, bw, bh+1 );
        src2 = get_ref( pix[1], &stride, m->p_fref, m->i_stride[0], omx-2, omy, bw+4, bh );
        src1 = src0 + stride;
        src3 = src2 + 1;
        x264_pixel_sad_x4_16x16( m->p_fenc[0], src0, src1, src2, src3, stride, costs );
        COPY2_IF_LT( bcost, costs[0] + p_cost_mvx[omx  ] + p_cost_mvy[omy-2], bmy, omy-2 );
        COPY2_IF_LT( bcost, costs[1] + p_cost_mvx[omx  ] + p_cost_mvy[omy+2], bmy, omy+2 );
        COPY3_IF_LT( bcost, costs[2] + p_cost_mvx[omx-2] + p_cost_mvy[omy  ], bmx, omx-2, bmy, omy );
        COPY3_IF_LT( bcost, costs[3] + p_cost_mvx[omx+2] + p_cost_mvy[omy  ], bmx, omx+2, bmy, omy );
        if( (bmx == omx) & (bmy == omy) )
            break;
    }

    if( !b_refine_qpel )
    {
        /* check for mvrange */
        if( bmy > h->mb.mv_max_spel[1] )
            bmy = h->mb.mv_max_spel[1];
        bcost = COST_MAX;
        COST_MV_SATD( bmx, bmy, -1 );
    }

    /* early termination when examining multiple reference frames */
    if( p_halfpel_thresh )
    {
        if( (bcost*7)>>3 > *p_halfpel_thresh )
        {
            m->cost = bcost;
            m->mv[0] = bmx;
            m->mv[1] = bmy;
            // don't need cost_mv
            return;
        }
        else if( bcost < *p_halfpel_thresh )
            *p_halfpel_thresh = bcost;
    }

    /* quarterpel diamond search */
    bdir = -1;
    for( i = qpel_iters; i > 0; i-- )
    {
        odir = bdir;
        omx = bmx;
        omy = bmy;
        COST_MV_SATD( omx, omy - 1, 0 );
        COST_MV_SATD( omx, omy + 1, 1 );
        COST_MV_SATD( omx - 1, omy, 2 );
        COST_MV_SATD( omx + 1, omy, 3 );
        if( bmx == omx && bmy == omy )
            break;
    }

    /* check for mvrange */
    if( bmy > h->mb.mv_max_spel[1] )
    {
        bmy = h->mb.mv_max_spel[1];
        bcost = COST_MAX;
        COST_MV_SATD( bmx, bmy, -1 );
    }

    m->cost = bcost;
    m->mv[0] = bmx;
    m->mv[1] = bmy;
    m->cost_mv = p_cost_mvx[ bmx ] + p_cost_mvy[ bmy ];
}

static inline int x264_clip3( int v, int i_min, int i_max )
{
    return ( (v < i_min) ? i_min : (v > i_max) ? i_max : v );
}

// -- Original function pointers hard coded

static uint8_t *get_ref( uint8_t *dst, int *i_dst_stride, uint8_t *src[4], int i_src_stride,
                         int mvx, int mvy, int i_width, int i_height )
{
    int qpel_idx = ((mvy&3)<<2) + (mvx&3);
    int offset = (mvy>>2)*i_src_stride + (mvx>>2);
    uint8_t *src1 = src[hpel_ref0[qpel_idx]] + offset + ((mvy&3) == 3) * i_src_stride;

    /*if( qpel_idx & 5 ) /* qpel interpolation needed */
    /*{
        uint8_t *src2 = src[hpel_ref1[qpel_idx]] + offset + ((mvx&3) == 3);
        pixel_avg( dst, *i_dst_stride, src1, i_src_stride,
                   src2, i_src_stride, i_width, i_height );
        return dst;
    }
    else*/
    {
        *i_dst_stride = i_src_stride;
        return src1;
    }

}

static inline void pixel_avg( uint8_t *dst,  int i_dst_stride, uint8_t *ppu_src1, int i_src1_stride,
                              uint8_t *ppu_src2, int i_src2_stride, int i_width, int i_height )
{
    int x, y;
    uint8_t *src1 = x264_malloc(i_width);
    uint8_t *src2 = x264_malloc(i_width);

    for( y = 0; y < i_height; y++ )
    {
        dma_transfer(src1, ppu_src1, sizeof(uint8_t)*i_width, 0);
        dma_transfer(src2, ppu_src2, sizeof(uint8_t)*i_width, 0);
        
        for( x = 0; x < i_width; x++ )
        {
            dst[x] = ( src1[x] + src2[x] + 1 ) >> 1;
        }
        dst  += i_dst_stride;
        ppu_src1 += i_src1_stride;
        ppu_src2 += i_src2_stride;
    }
    x264_free(src1);
    x264_free(src2);
}

static void mc_chroma( uint8_t *dst, int i_dst_stride, uint8_t *src, int i_src_stride,
                       int mvx, int mvy, int i_width, int i_height )
{
    uint8_t *srcp;
    int x, y;

    const int d8x = mvx&0x07;
    const int d8y = mvy&0x07;

    const int cA = (8-d8x)*(8-d8y);
    const int cB = d8x    *(8-d8y);
    const int cC = (8-d8x)*d8y;
    const int cD = d8x    *d8y;

    src  += (mvy >> 3) * i_src_stride + (mvx >> 3);
    srcp = &src[i_src_stride];

    for( y = 0; y < i_height; y++ )
    {
        for( x = 0; x < i_width; x++ )
        {
            dst[x] = ( cA*src[x]  + cB*src[x+1] +
                       cC*srcp[x] + cD*srcp[x+1] + 32 ) >> 6;
        }
        dst  += i_dst_stride;

        src   = srcp;
        srcp += i_src_stride;
    }
}

/****************************************************************************
 * pixel_sad_WxH
 ****************************************************************************/
#define PIXEL_SAD_C( name, lx, ly ) \
static int name( uint8_t *ppu_pix1, int i_stride_pix1,  \
                 uint8_t *ppu_pix2, int i_stride_pix2 ) \
{                                                   \
    int i_sum = 0;                                  \
    int x, y;                                       \
    \
    unsigned int spu_start, spu_stop1;              \
    spu_write_decrementer(0x7fffffff);              \
    spu_start = spu_read_decrementer();             \
    \
    \
    \
    uint8_t  pix1_arr[16 * ly], *pix1, pix2[lx];     \
    pix1 = pix1_arr;                                   \
    dma_transfer(pix1_arr, (unsigned int)ppu_pix1, 16*ly, 0);\
    \
    for( y = 0; y < ly; y++ )                       \
    {                                               \
        dma_transfer(pix2, (unsigned int)ppu_pix2, sizeof(uint8_t)*lx, 0);\
        for( x = 0; x < lx; x++ )                   \
        {                                           \
            i_sum += abs( pix1[x] - pix2[x] );      \
        }                                           \
        pix1 += 16;                                 \
        ppu_pix1 += i_stride_pix1;                      \
        ppu_pix2 += i_stride_pix2;                      \
    }                                               \
    \
    spu_stop1 = spu_read_decrementer();\
    \
    sad_time += (spu_start - spu_stop1);\
    \
    return i_sum;                                   \
}


PIXEL_SAD_C( x264_pixel_sad_16x16, 16, 16 )
PIXEL_SAD_C( x264_pixel_sad_16x8,  16,  8 )
PIXEL_SAD_C( x264_pixel_sad_8x16,   8, 16 )
PIXEL_SAD_C( x264_pixel_sad_8x8,    8,  8 )
PIXEL_SAD_C( x264_pixel_sad_8x4,    8,  4 )
PIXEL_SAD_C( x264_pixel_sad_4x8,    4,  8 )
PIXEL_SAD_C( x264_pixel_sad_4x4,    4,  4 )
//****************************************************************************

static void x264_pixel_sad_x3_16x16( uint8_t *fenc, uint8_t *pix0, uint8_t *pix1, uint8_t *pix2, int i_stride, int scores[3] )
{
    scores[0] = x264_pixel_sad_16x16( fenc, FENC_STRIDE, pix0, i_stride );
    scores[1] = x264_pixel_sad_16x16( fenc, FENC_STRIDE, pix1, i_stride );
    scores[2] = x264_pixel_sad_16x16( fenc, FENC_STRIDE, pix2, i_stride );
}

static void x264_pixel_sad_x4_16x16( uint8_t *fenc, uint8_t *pix0, uint8_t *pix1, uint8_t *pix2, uint8_t *pix3, int i_stride, int scores[4] )
{
    scores[0] = x264_pixel_sad_16x16( fenc, FENC_STRIDE, pix0, i_stride );
    scores[1] = x264_pixel_sad_16x16( fenc, FENC_STRIDE, pix1, i_stride );
    scores[2] = x264_pixel_sad_16x16( fenc, FENC_STRIDE, pix2, i_stride );
    scores[3] = x264_pixel_sad_16x16( fenc, FENC_STRIDE, pix3, i_stride );
}

static void x264_pixel_sad_x4_16x8( uint8_t *fenc, uint8_t *pix0, uint8_t *pix1, uint8_t *pix2, uint8_t *pix3, int i_stride, int scores[4] )
{
    scores[0] = x264_pixel_sad_16x8( fenc, FENC_STRIDE, pix0, i_stride );
    scores[1] = x264_pixel_sad_16x8( fenc, FENC_STRIDE, pix1, i_stride );
    scores[2] = x264_pixel_sad_16x8( fenc, FENC_STRIDE, pix2, i_stride );
    scores[3] = x264_pixel_sad_16x8( fenc, FENC_STRIDE, pix3, i_stride );
}

static void x264_pixel_sad_x4_8x16( uint8_t *fenc, uint8_t *pix0, uint8_t *pix1, uint8_t *pix2, uint8_t *pix3, int i_stride, int scores[4] )
{
    scores[0] = x264_pixel_sad_8x16( fenc, FENC_STRIDE, pix0, i_stride );
    scores[1] = x264_pixel_sad_8x16( fenc, FENC_STRIDE, pix1, i_stride );
    scores[2] = x264_pixel_sad_8x16( fenc, FENC_STRIDE, pix2, i_stride );
    scores[3] = x264_pixel_sad_8x16( fenc, FENC_STRIDE, pix3, i_stride );
}

static void x264_pixel_sad_x4_8x8( uint8_t *fenc, uint8_t *pix0, uint8_t *pix1, uint8_t *pix2, uint8_t *pix3, int i_stride, int scores[4] )
{
    scores[0] = x264_pixel_sad_8x8( fenc, FENC_STRIDE, pix0, i_stride );
    scores[1] = x264_pixel_sad_8x8( fenc, FENC_STRIDE, pix1, i_stride );
    scores[2] = x264_pixel_sad_8x8( fenc, FENC_STRIDE, pix2, i_stride );
    scores[3] = x264_pixel_sad_8x8( fenc, FENC_STRIDE, pix3, i_stride );
}

static void x264_pixel_sad_x4_8x4( uint8_t *fenc, uint8_t *pix0, uint8_t *pix1, uint8_t *pix2, uint8_t *pix3, int i_stride, int scores[4] )
{
    scores[0] = x264_pixel_sad_8x4( fenc, FENC_STRIDE, pix0, i_stride );
    scores[1] = x264_pixel_sad_8x4( fenc, FENC_STRIDE, pix1, i_stride );
    scores[2] = x264_pixel_sad_8x4( fenc, FENC_STRIDE, pix2, i_stride );
    scores[3] = x264_pixel_sad_8x4( fenc, FENC_STRIDE, pix3, i_stride );
}

static void x264_pixel_sad_x4_4x8( uint8_t *fenc, uint8_t *pix0, uint8_t *pix1, uint8_t *pix2, uint8_t *pix3, int i_stride, int scores[4] )
{
    scores[0] = x264_pixel_sad_4x8( fenc, FENC_STRIDE, pix0, i_stride );
    scores[1] = x264_pixel_sad_4x8( fenc, FENC_STRIDE, pix1, i_stride );
    scores[2] = x264_pixel_sad_4x8( fenc, FENC_STRIDE, pix2, i_stride );
    scores[3] = x264_pixel_sad_4x8( fenc, FENC_STRIDE, pix3, i_stride );
}

static void x264_pixel_sad_x4_4x4( uint8_t *fenc, uint8_t *pix0, uint8_t *pix1, uint8_t *pix2, uint8_t *pix3, int i_stride, int scores[4] )
{
    scores[0] = x264_pixel_sad_4x4( fenc, FENC_STRIDE, pix0, i_stride );
    scores[1] = x264_pixel_sad_4x4( fenc, FENC_STRIDE, pix1, i_stride );
    scores[2] = x264_pixel_sad_4x4( fenc, FENC_STRIDE, pix2, i_stride );
    scores[3] = x264_pixel_sad_4x4( fenc, FENC_STRIDE, pix3, i_stride );
}


