#ifndef __ME_SPU_H__
#define __ME_SPU_H__

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <spu_mfcio.h>

#include "common.h"

#define BIGENDIAN  1
#define COST_MAX  (1<<28)

#define X264_ME_DIA  0
#define X264_ME_HEX  1
#define X264_ME_UMH  2
#define X264_ME_ESA  3
#define X264_ME_TESA 4

static inline int x264_clip3( int, int, int);
static uint8_t *get_ref( uint8_t *, int *, uint8_t **, int, int, int, int, int);
static inline void pixel_avg( uint8_t *,  int, uint8_t *, int, uint8_t *, int, int, int);
static void mc_chroma( uint8_t *, int, uint8_t *, int, int, int, int, int);
static void x264_pixel_sad_x3_16x16( uint8_t *, uint8_t *, uint8_t *, uint8_t *, int, int[]);
void x264_me_search_ref(x264_t *, x264_me_t *, int16_t (*)[], int, int *, int);
static void refine_subpel( x264_t *, x264_me_t *, int, int, int *, int);

static int x264_pixel_sad_16x16( uint8_t *, int, uint8_t *, int);
static int x264_pixel_sad_16x8( uint8_t *, int, uint8_t *, int);
static int x264_pixel_sad_8x16( uint8_t *, int, uint8_t *, int);
static int x264_pixel_sad_8x8( uint8_t *, int, uint8_t *, int);
static int x264_pixel_sad_8x4( uint8_t *, int, uint8_t *, int);
static int x264_pixel_sad_4x8( uint8_t *, int, uint8_t *, int);
static int x264_pixel_sad_4x4( uint8_t *, int, uint8_t *, int);

static void x264_pixel_sad_x4_16x16( uint8_t *, uint8_t *, uint8_t *, uint8_t *, uint8_t *, int, int[] );
static void x264_pixel_sad_x4_16x8( uint8_t *, uint8_t *, uint8_t *, uint8_t *, uint8_t *, int, int[] );
static void x264_pixel_sad_x4_8x16( uint8_t *, uint8_t *, uint8_t *, uint8_t *, uint8_t *, int, int[] );
static void x264_pixel_sad_x4_8x8( uint8_t *, uint8_t *, uint8_t *, uint8_t *, uint8_t *, int, int[] );
static void x264_pixel_sad_x4_8x4( uint8_t *, uint8_t *, uint8_t *, uint8_t *, uint8_t *, int, int[] );
static void x264_pixel_sad_x4_4x8( uint8_t *, uint8_t *, uint8_t *, uint8_t *, uint8_t *, int, int[] );
static void x264_pixel_sad_x4_4x4( uint8_t *, uint8_t *, uint8_t *, uint8_t *, uint8_t *, int, int[] );


//void x264_me_search_ref( struct x264_t *);

// --- Bigendian vs Little Endian

static inline uint32_t pack16to32_mask( int a, int b )
{
    #ifdef WORDS_BIGENDIAN
       return (b&0xFFFF) + (a<<16);
    #else
       return (a&0xFFFF) + (b<<16);
    #endif
}

// -- Comparisons ----------------------------

#define COPY2_IF_LT(x,y,a,b)\
if((y)<(x))\
{\
    (x)=(y);\
    (a)=(b);\
}

#define COPY3_IF_LT(x,y,a,b,c,d)\
if((y)<(x))\
{\
    (x)=(y);\
    (a)=(b);\
    (c)=(d);\
}

// --- Motion Vector Comparison ---------------

#define COST_MV( mx, my )\
{\
    dma_transfer(&cost_mvx, (unsigned)&p_cost_mvx[(mx)<<2], sizeof(int16_t), 0);\
    dma_transfer(&cost_mvy, (unsigned)&p_cost_mvy[(my)<<2], sizeof(int16_t), 0);\
    int cost;\
    switch(i_pixel) \
    { \
        case 0:\
            cost = x264_pixel_sad_16x16( m.p_fenc[0], FENC_STRIDE, &p_fref[(my)*m.i_stride[0]+(mx)], m.i_stride[0] ) \
                       + cost_mvx + cost_mvy; \
            break;\
        case 1:\
            cost = x264_pixel_sad_16x8( m.p_fenc[0], FENC_STRIDE, &p_fref[(my)*m.i_stride[0]+(mx)], m.i_stride[0] ) \
                       + cost_mvx + cost_mvy; \
            break;\
        case 2:\
            cost = x264_pixel_sad_8x16( m.p_fenc[0], FENC_STRIDE, &p_fref[(my)*m.i_stride[0]+(mx)], m.i_stride[0] ) \
                       + cost_mvx + cost_mvy; \
            break;\
        case 3:\
            cost = x264_pixel_sad_8x8( m.p_fenc[0], FENC_STRIDE, &p_fref[(my)*m.i_stride[0]+(mx)], m.i_stride[0] ) \
                       + cost_mvx + cost_mvy; \
            break;\
        case 4:\
            cost = x264_pixel_sad_8x4( m.p_fenc[0], FENC_STRIDE, &p_fref[(my)*m.i_stride[0]+(mx)], m.i_stride[0] ) \
                       + cost_mvx + cost_mvy; \
            break;\
        case 5:\
            cost = x264_pixel_sad_4x8( m.p_fenc[0], FENC_STRIDE, &p_fref[(my)*m.i_stride[0]+(mx)], m.i_stride[0] ) \
                       + cost_mvx + cost_mvy; \
            break;\
        case 6:\
            cost = x264_pixel_sad_4x4( m.p_fenc[0], FENC_STRIDE, &p_fref[(my)*m.i_stride[0]+(mx)], m.i_stride[0] ) \
                       + cost_mvx + cost_mvy; \
            break;\                                                         
    } \
     \            
    COPY3_IF_LT( bcost, cost, bmx, mx, bmy, my );\
}

#define COST_MV_SAD( mx, my ) \
{ \
    int stride = 16; \
    uint8_t *src = get_ref( pix[0], &stride, m->p_fref, m->i_stride[0], mx, my, bw, bh ); \
    int cost = x264_pixel_sad_16x16( m->p_fenc[0], FENC_STRIDE, src, stride ) \
             + p_cost_mvx[ mx ] + p_cost_mvy[ my ]; \
    COPY3_IF_LT( bcost, cost, bmx, mx, bmy, my ); \
}

#define COST_MV_SATD( mx, my, dir ) \
if( b_refine_qpel || (dir^1) != odir ) \
{ \
    int stride = 16; \
    uint8_t *src = get_ref( pix[0], &stride, m->p_fref, m->i_stride[0], mx, my, bw, bh ); \
    int cost = x264_pixel_sad_16x16( m->p_fenc[0], FENC_STRIDE, src, stride ) + p_cost_mvx[ mx ] + p_cost_mvy[ my ]; \
    if( b_chroma_me && cost < bcost ) \
    { \
        mc_chroma( pix[0], 8, m->p_fref[4], m->i_stride[1], mx, my, bw/2, bh/2 ); \
        cost += x264_pixel_sad_8x8( m->p_fenc[1], FENC_STRIDE, pix[0], 8 ); \
        if( cost < bcost ) \
        { \
            mc_chroma( pix[0], 8, m->p_fref[5], m->i_stride[1], mx, my, bw/2, bh/2 ); \
            cost += x264_pixel_sad_8x8( m->p_fenc[2], FENC_STRIDE, pix[0], 8 ); \
        } \
    } \
    if( cost < bcost ) \
    {                  \
        bcost = cost;  \
        bmx = mx;      \
        bmy = my;      \
        bdir = dir;    \
    } \
}

#define COST_MV_HPEL( mx, my ) \
{ \
    int stride = 16; \
    uint8_t *src = get_ref( pix, &stride, m.p_fref, m.i_stride[0], mx, my, bw, bh ); \
    dma_transfer(&cost_mvx, (unsigned int)p_cost_mvx + mx, sizeof(int16_t), 0);\
    dma_transfer(&cost_mvy, (unsigned int)p_cost_mvy + my, sizeof(int16_t), 0);\
    int cost;\
    switch(i_pixel) \
    { \
        case 0:\
            cost = x264_pixel_sad_16x16( m.p_fenc[0], FENC_STRIDE, src, stride ) \
                       + cost_mvx + cost_mvy; \
            break;\
        case 1:\
            cost = x264_pixel_sad_16x8( m.p_fenc[0], FENC_STRIDE, src, stride ) \
                       + cost_mvx + cost_mvy; \
            break;\
        case 2:\
            cost = x264_pixel_sad_8x16( m.p_fenc[0], FENC_STRIDE, src, stride ) \
                       + cost_mvx + cost_mvy; \
            break;\
        case 3:\
            cost = x264_pixel_sad_8x8( m.p_fenc[0], FENC_STRIDE, src, stride ) \
                       + cost_mvx + cost_mvy; \
            break;\
        case 4:\
            cost = x264_pixel_sad_8x4( m.p_fenc[0], FENC_STRIDE, src, stride ) \
                       + cost_mvx + cost_mvy; \
            break;\
        case 5:\
            cost = x264_pixel_sad_4x8( m.p_fenc[0], FENC_STRIDE, src, stride ) \
                       + cost_mvx + cost_mvy; \
            break;\
        case 6:\
            cost = x264_pixel_sad_4x4( m.p_fenc[0], FENC_STRIDE, src, stride ) \
                       + cost_mvx + cost_mvy; \
            break;\                                                         
    } \
    COPY3_IF_LT( bpred_cost, cost, bpred_mx, mx, bpred_my, my ); \
}

#define COST_MV_X3_DIR( m0x, m0y, m1x, m1y, m2x, m2y, costs )\
{\
    uint8_t *pix_base = p_fref + bmx + bmy*m->i_stride[0];\
    x264_pixel_sad_x3_16x16( m->p_fenc[0],\
        pix_base + (m0x) + (m0y)*m->i_stride[0],\
        pix_base + (m1x) + (m1y)*m->i_stride[0],\
        pix_base + (m2x) + (m2y)*m->i_stride[0],\
        m->i_stride[0], costs );\
    (costs)[0] += BITS_MVD( bmx+(m0x), bmy+(m0y) );\
    (costs)[1] += BITS_MVD( bmx+(m1x), bmy+(m1y) );\
    (costs)[2] += BITS_MVD( bmx+(m2x), bmy+(m2y) );\
}

#define COST_MV_X4( m0x, m0y, m1x, m1y, m2x, m2y, m3x, m3y )\
{\
    uint8_t *pix_base = p_fref + omx + omy*m.i_stride[0];\
    switch(i_pixel) \
    { \
        case 0:\
            x264_pixel_sad_x4_16x16( m.p_fenc[0],\
                   pix_base + (m0x) + (m0y)*m.i_stride[0],\
                   pix_base + (m1x) + (m1y)*m.i_stride[0],\
                   pix_base + (m2x) + (m2y)*m.i_stride[0],\
                   pix_base + (m3x) + (m3y)*m.i_stride[0],\
                   m.i_stride[0], costs );\
            break;\
        case 1:\
            x264_pixel_sad_x4_16x8( m.p_fenc[0],\
                   pix_base + (m0x) + (m0y)*m.i_stride[0],\
                   pix_base + (m1x) + (m1y)*m.i_stride[0],\
                   pix_base + (m2x) + (m2y)*m.i_stride[0],\
                   pix_base + (m3x) + (m3y)*m.i_stride[0],\
                   m.i_stride[0], costs );\
            break;\
        case 2:\
            x264_pixel_sad_x4_8x16( m.p_fenc[0],\
                   pix_base + (m0x) + (m0y)*m.i_stride[0],\
                   pix_base + (m1x) + (m1y)*m.i_stride[0],\
                   pix_base + (m2x) + (m2y)*m.i_stride[0],\
                   pix_base + (m3x) + (m3y)*m.i_stride[0],\
                   m.i_stride[0], costs );\
            break;\
        case 3:\
            x264_pixel_sad_x4_8x8( m.p_fenc[0],\
                   pix_base + (m0x) + (m0y)*m.i_stride[0],\
                   pix_base + (m1x) + (m1y)*m.i_stride[0],\
                   pix_base + (m2x) + (m2y)*m.i_stride[0],\
                   pix_base + (m3x) + (m3y)*m.i_stride[0],\
                   m.i_stride[0], costs );\
            break;\
        case 4:\
            x264_pixel_sad_x4_8x4( m.p_fenc[0],\
                   pix_base + (m0x) + (m0y)*m.i_stride[0],\
                   pix_base + (m1x) + (m1y)*m.i_stride[0],\
                   pix_base + (m2x) + (m2y)*m.i_stride[0],\
                   pix_base + (m3x) + (m3y)*m.i_stride[0],\
                   m.i_stride[0], costs );\
            break;\
        case 5:\
            x264_pixel_sad_x4_4x8( m.p_fenc[0],\
                   pix_base + (m0x) + (m0y)*m.i_stride[0],\
                   pix_base + (m1x) + (m1y)*m.i_stride[0],\
                   pix_base + (m2x) + (m2y)*m.i_stride[0],\
                   pix_base + (m3x) + (m3y)*m.i_stride[0],\
                   m.i_stride[0], costs );\
            break;\
        case 6:\
            x264_pixel_sad_x4_4x4( m.p_fenc[0],\
                   pix_base + (m0x) + (m0y)*m.i_stride[0],\
                   pix_base + (m1x) + (m1y)*m.i_stride[0],\
                   pix_base + (m2x) + (m2y)*m.i_stride[0],\
                   pix_base + (m3x) + (m3y)*m.i_stride[0],\
                   m.i_stride[0], costs );\
            break;\                                                         
    } \
    costs[0] += BITS_MVD( omx+(m0x), omy+(m0y) );\
    costs[1] += BITS_MVD( omx+(m1x), omy+(m1y) );\
    costs[2] += BITS_MVD( omx+(m2x), omy+(m2y) );\
    costs[3] += BITS_MVD( omx+(m3x), omy+(m3y) );\
    COPY3_IF_LT( bcost, costs[0], bmx, omx+(m0x), bmy, omy+(m0y) );\
    COPY3_IF_LT( bcost, costs[1], bmx, omx+(m1x), bmy, omy+(m1y) );\
    COPY3_IF_LT( bcost, costs[2], bmx, omx+(m2x), bmy, omy+(m2y) );\
    COPY3_IF_LT( bcost, costs[3], bmx, omx+(m3x), bmy, omy+(m3y) );\
}

// ---

#define BITS_MVD( mx, my )    (get_int16_from_ppu(&p_cost_mvx[(mx)<<2]) + get_int16_from_ppu(&p_cost_mvy[(my)<<2]))

// ---
    
#endif
